#
# Copyright 2006, 2007 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use, modify,
# copy, or redistribute it subject to the terms and conditions of the GNU
# General Public License v.2.  This program is distributed in the hope that it
# will be useful, but WITHOUT ANY WARRANTY expressed or implied, including the
# implied warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Red Hat
# trademarks that are incorporated in the source code or documentation are not
# subject to the GNU General Public License and may only be used or replicated
# with the express permission of Red Hat, Inc.
#

import os, sys
import optparse
import traceback
import rpmUtils.arch
import snack

from rhpl.translate import _, N_
from string import joinfields

import snake.log
import snake.client
import snake.install
from snake.constants import *

# import avahi
try:
    import snake.zeroconf
except ImportError:
    has_zeroconf = False
    pass
else:
    has_zeroconf = True

# setup log handler
log = snake.log.setup_logger()

# define the sorted() function for older python
if sys.version_info < (2,4):
    def sorted(iterable, cmp=None, key=None, reverse=False):
        "return a sorted copy of its input"
        seq = list(iterable)
        if reverse:
            seq.reverse()        # preserve stability
        if key is not None:
            seq = [(key(elem), i, elem) for i, elem in enumerate(seq)]
        seq.sort(cmp)
        if key is not None:
            seq = [elem for (key, i, elem) in seq]
        if reverse:
            seq.reverse()
        return seq

def formatException (type, value, tb):
    lst = traceback.format_tb(tb)
    # Reverse the order that tracebacks are printed
    #lst.reverse() # this makes it too confusing for me (jlaska)
    lst.insert(0, "Snake exception report\n")
    lst.insert(1, 'Traceback (most recent call first):\n')
    lst.extend(traceback.format_exception_only(type, value))
    return lst

def handleException(screen, (type, value, tb)):
    # restore original exception handler
    sys.excepthook = sys.__excepthook__

    # get traceback information
    list = formatException (type, value, tb)
    text = joinfields (list, "")

    # save to local storage first
    out = open("/tmp/snakedump.txt", "w")
    #dumpException (out, text, tb, anaconda)
    out.write(text)
    out.close()

    mainWin = MainExceptionWindow(screen, shortTraceback=text)
    rc = mainWin.run()

    # not sure why this doesn't call __del__ properly
    # del mainWin
    screen.finish()

    log.critical(text)
    return rc

class WaitWindow:
    '''taken almost directly from anaconda/text.py'''
    def pop(self):
        self.screen.popWindow()
        self.screen.refresh()

    def __init__(self, screen, title, text, width):
        self.screen = screen
        if (len(text) < width): width = len(text)

        t = snack.TextboxReflowed(width, text)

        g = snack.GridForm(self.screen, title, 1, 1)
        g.add(t, 0, 0)
        g.draw()
        self.screen.refresh()

def waitWindow(screen, title, text, width):
    return WaitWindow(screen, title, text, width)

class SnakeWindow:
    def __init__(self, screen, tui, cfg):
        self.screen = screen
        self.tui = tui
        self.cfg = cfg
        size = snack._snack.size()

class MainExceptionWindow:
    def __init__(self, screen, shortTraceback):
        self.screen = screen
        self.text = "%s\n\n" % shortTraceback

    def run(self):
        #rc = snack.ButtonChoiceWindow(self.screen, _("Exception Occurred"),
        #                              exceptionText, [TEXT_EXIT_BUTTON], self.screen.width-20)

        bb = snack.ButtonBar(self.screen, [TEXT_EXIT_BUTTON])
        t1 = snack.TextboxReflowed(self.screen.width-20, exceptionText, maxHeight = self.screen.height - 16)
        t2 = snack.TextboxReflowed(self.screen.width-20, self.text, maxHeight = self.screen.height - 16)

        g = snack.GridFormHelp(self.screen, _("Exception Occurred"), None, 1, 3)
        g.add(t1, 0, 0, padding = (0, 0, 0, 1))
        g.add(t2, 0, 1, padding = (0, 0, 0, 1))
        g.add(bb, 0, 2, growx = 1)
        rc = bb.buttonPressed(g.runOnce(None, None))

        return TEXT_EXIT_CHECK

    def __del_(self):
        if self.screen:
            self.screen.finish()

class FindServerWindow(SnakeWindow):
    def __call__(self):

        # if no server specified, try to find using zeroconf
        if not self.cfg.server and has_zeroconf:
            '''attempt to find a local snake-server using avahi discovery'''
            window = waitWindow(self.screen, "Searching Network", "Searching your local network for a snake server.", self.tui.width)
            try:
                avahi_client = snake.zeroconf.AvahiClient()
                self.cfg.server = avahi_client.get_server()
            except:
                pass
            window.pop() # remove popup when found

        # No server found
        if not self.cfg.server:
            '''no server found, return with error'''
            error_message = "Could not find a snake server. Try using the -s option."
            rc = snack.ButtonChoiceWindow(self.screen,"No Server Found", error_message,
                    [TEXT_EXIT_BUTTON], self.tui.width)
            return TEXT_EXIT_CHECK

        # TODO: server should be just a hostname ... use optparse to get --port
        # We were probably just given a hostname. Make it a proper URL.
        if not ':' in self.cfg.server:
            self.cfg.server = 'http://%s:2903' % self.cfg.server

        # Check that the server is alive
        if not snake.client.check_server(self.cfg.server):
            '''the server is not responding'''
            error_message = "The snake server %s is not responding. " \
                            "Please check your server URL and try again." % self.cfg.server
            r = snack.ButtonChoiceWindow(self.screen,"Unable To Connect", error_message,
                    [TEXT_EXIT_BUTTON], self.tui.width)
            return TEXT_EXIT_CHECK

        # Connect and save connection object
        self.tui.xmlrpc = snake.client.connect(self.cfg.server)

        return TEXT_OK_CHECK

class TreeChoiceWindow(SnakeWindow):
    def __init__(self, *kwargs):
        SnakeWindow.__init__(self, *kwargs)
        self.cfg.tree = None

    def __call__(self):

        args = {"arch": self.cfg.arch}
        treedict = snake.client.treedict(self.tui.xmlrpc, **args)

        if not treedict:
            error_message = "No %s trees found on the snake server %s" % (self.cfg.arch, self.cfg.server)
            r = snack.ButtonChoiceWindow(screen,"No Trees Found", error_message,
                    [TEXT_EXIT_BUTTON], width)
            return TEXT_EXIT_CHECK

        items = [(str(treedict[k]),k) for k in sorted(treedict.keys())]
        (button, id) = snack.ListboxChoiceWindow(self.screen, "Choose a tree", "What tree would you like to install?",
            items, width=self.tui.width, height=min(self.tui.height,len(items)),
            buttons=[TEXT_OK_BUTTON, TEXT_CANCEL_BUTTON], scroll=1)

        if button is None:
            button = TEXT_OK_CHECK

        self.cfg.tree = treedict[id]

        return button

class URIChoiceWindow(SnakeWindow):
    def __init__(self, *kwargs):
        SnakeWindow.__init__(self, *kwargs)
        self.cfg.uri = None

    def __call__(self):

        # TODO sort usefully?
        urilist = [snake.uri.parse(u) for u in self.cfg.tree.uris]

        #FIXME - needs to take into account long url's and small screens
        # nfs uri objects do not have a hostname member :(
        items = [("%5s - %s:%s" % (u.scheme, \
            u.hostname or u.path.split(':',1)[0], \
            u.hostname and u.path or u.path.split(':',1)[1]), u.geturl()) for u in urilist]

        (button, self.cfg.uri) = snack.ListboxChoiceWindow(self.screen, "Choose a mirror",
            "Which mirror would you like to install from?",
            items, width=self.tui.width, height=min(self.tui.height,len(items)),
            buttons=[TEXT_OK_BUTTON, TEXT_BACK_BUTTON, TEXT_CANCEL_BUTTON], scroll=1)

        if button is None:
            button = TEXT_OK_CHECK

        return button

class KSChoiceWindow(SnakeWindow):
    def __init__(self, *kwargs):
        SnakeWindow.__init__(self, *kwargs)
        self.cfg.ks = None

    def __call__(self):

        kslist = self.tui.xmlrpc.kickstart.list()
        items = [('Manual install (no kickstart)',None)]
        items += [("%s - %s"%(name,desc), name) for name,desc in kslist]

        (button, self.cfg.ks) = snack.ListboxChoiceWindow(self.screen, "Select an installation method",
            "Here, you may choose a manual installation or select from a list of available kickstart installations.",
            items, width=self.tui.width, height=min(self.tui.height,len(items)),
            buttons=[TEXT_OK_BUTTON, TEXT_BACK_BUTTON, TEXT_CANCEL_BUTTON], scroll=1)

        if button is None:
            button = TEXT_OK_CHECK

        return button

class ReviewWindow(SnakeWindow):
    def __call__(self):

        confirm_msg = '''You are about to configure your system to install this tree:
  %s
which will be installed from the following location:
  %s
'''% (str(self.cfg.tree),self.cfg.uri)
        if self.cfg.ks: confirm_msg += "using the following kickstart:\n  %s\n" % self.cfg.ks
        confirm_msg += "\nAre you sure you wish to proceed?"

        rc = snack.ButtonChoiceWindow(self.screen, "Confirm selections?", confirm_msg,
                [TEXT_OK_BUTTON, TEXT_BACK_BUTTON, TEXT_CANCEL_BUTTON], self.tui.width)

        # handle when F12 is pressed
        if rc is None:
            rc = TEXT_OK_CHECK

        # Setup bootloader?
        # FIXME: verify selections and present error if needed
        if rc in [TEXT_OK_CHECK,TEXT_F12_CHECK]:

            fpargs = dict()

            # Are any arch overrides needed?
            # FIXME: this sort of monkeying with arch should really happen in the 
            # client libs, rather than each app..
            from snake.machineinfo import isParaVirt
            if isParaVirt():
                '''paravirt guests will need xen boot images'''
                fpargs["arch"] = "xen"
            else:
                fpargs["arch"] = self.cfg.arch
            if fpargs["arch"] == 'ppc' and os.uname()[4] == 'ppc64':
                fpargs["arch"] = 'ppc64'

            # Generate a kickstart file
            ksdata = None
            if self.cfg.ks:
                # set ver to something usable by pykickstart
                ver = 'devel'
                if self.cfg.tree.version.isdigit():
                    ver = "%s %s" % (self.cfg.tree.family, self.cfg.tree.version)

                u = snake.uri.parse(self.cfg.uri)
                method = ""
                if u.scheme in ("https","http","ftp"):
                    method = "url --url=%s" % u.geturl()
                elif u.scheme == "nfs":
                    method = "nfs --server=%s --dir=%s" % tuple(u.path.split(':',1))
                else:
                    # TODO: harddrive, cdrom support
                    error_message = "The URI '%s' is not recognized as an anaconda installation source" % self.cfg.uri
                    r = snack.ButtonChoiceWindow(screen,"Unrecognized URI", error_message,
                            exit_button, def_width)
                    return (r, "Cancelled.")
                # FIXME: need to emit an error message and drop 'ver' if it turns out
                # we can't generate a kickstart for that version (or if 'ver' is 
                # incomprehensible)
                fpargs["ksdata"] = self.tui.xmlrpc.kickstart.generate(self.cfg.ks,{'commands':[method],'version':ver})

                # was a kickstart method of delivery specified?
                if self.cfg.ksmethod:
                    fpargs["ksmethod"] = self.cfg.ksmethod

            # setup status callback
            global status_window
            status_window = None
            def print_status(msg):
                '''Display popup window displaying current fetch_and_prep status'''
                global status_window
                if status_window != None: status_window.pop()
                status_window = waitWindow(self.screen, "Status", msg, self.tui.width)
            fpargs["status_callback"] = print_status

            # Give fetch_and_prep the Tree object
            fpargs["tree"] = self.cfg.tree

            if snake.install.fetch_and_prep(self.cfg.uri, **fpargs):
                if status_window != None: status_window.pop()
                return TEXT_OK_CHECK
            else:
                # FIXME: need improved return code handling
                error_message = _("An error occurred while setting up the installer.")
                r = snack.ButtonChoiceWindow(self.screen,"Error", error_message,
                        [TEXT_EXIT_BUTTON], self.tui.width)
                return TEXT_EXIT_CHECK

        return rc

class RebootWindow(SnakeWindow):
    def __call__(self):

        confirm_msg = _("Congratulations, your system is now ready to install. "
                      "The installation will start when your system reboots. \n\n"
                      "Would you like to reboot now?")

        rc = snack.ButtonChoiceWindow(self.screen,"Reboot?", confirm_msg,
                [TEXT_OK_BUTTON, TEXT_EXIT_BUTTON], self.tui.width)

        # Reboot is handled after the snack ui is cleaned up
        #if r in [TEXT_YES_CHECK,TEXT_OK_CHECK]:
        #    os.system("/sbin/reboot")

        return rc

class SnakeTui(object):

    def __init__(self, screen, opt):
        self.screen = screen
        self.size = snack._snack.size()

        # Pull default screen dimensions in to leave room for title/help/scroll elements
        self.width = min(self.size[0]-10,100)
        self.height = self.size[1]-15

        # FIXME: probably shouldn't reuse the OptionParser in this way
        self.cfg = opt

        self.xmlrpc = None # store the handle to the xmlrpc instance

        self.windows = [FindServerWindow,
                        TreeChoiceWindow,
                        URIChoiceWindow,
                        KSChoiceWindow,
                        ReviewWindow,
                        RebootWindow,
                       ]

        # setup an exception handler to display a formatted traceback
        sys.excepthook = lambda type, value, tb, screen=self.screen: handleException(screen, (type, value, tb))

    def run(self):
        # draw title
        self.screen.drawRootText(0, 0, "SNAKE Client on %s" % self.cfg.arch)

        # loop through ui windows
        step = 0
        while step < len(self.windows):
            win = self.windows[step](self.screen, self, self.cfg)
            rc = win()

            if rc == TEXT_BACK_CHECK:
                step = step - 1
            elif rc in [TEXT_OK_CHECK, TEXT_YES_CHECK, TEXT_F12_CHECK]:
                step = step + 1
            elif rc in [TEXT_EXIT_CHECK,TEXT_CANCEL_CHECK]:
                step = step + 1
                break

        # cleanup snack ui
        self.cleanup()

        # Provide a final status msg
        if step >= len(self.windows):
            if rc in [TEXT_OK_CHECK, TEXT_YES_CHECK, TEXT_F12_CHECK]:
                print "Done.  Your system will reboot to start the install."
                os.system("/sbin/reboot")
            else:
                print "Done.  You may now reboot your system to start the install."
        else:
            print "Cancelled."

        return rc

    def cleanup(self):
        if self.screen:
            self.screen.finish()

    def __del__(self):
        '''cleanup'''
        self.cleanup()

def setup_option_parser():
    parser=optparse.OptionParser()
    parser.add_option("-s", "--server",
                action="store", dest="server", default=os.environ.get("SNAKE_SERVER",""),
                help="URL of snake-server")
    parser.add_option("-a", "--arch",
                action="store", dest="arch", default=rpmUtils.arch.getBaseArch(),
                help="Only include trees of specified architecture (default: %default)")
    parser.add_option("-m", "--ksmethod",
                action="store", dest="ksmethod", default="initrd",
                help="Specify an alternative method for kickstart delivery (default: %default)")
    return parser

def main():
    parser = setup_option_parser()
    (opt,args) = parser.parse_args()

    screen = snack.SnackScreen()

    # TODO: make this a full-fledged SNAKE client (add/remove URIs, etc)
    tui = SnakeTui(screen, opt)
    rc = tui.run()
    sys.exit(rc)

if __name__ == '__main__':

    main()

